<?php
/**
 * A centralized place for data access for the application
 */
class Data {

    /**
     * Obtain a new connection to the database
     * @return SQLite3 A new connection to the database
     */
    private static function getConnection(): SQLite3 {
        $db = new SQLite3('../data/' . DATABASE_NAME);
        $db->exec('PRAGMA foreign_keys = ON;');
        return $db;
    }

    /**
     * Make sure the expected tables exist
     */
    public static function ensureDb(): void {
        $db = self::getConnection();
        $tables = array();
        $tableQuery = $db->query("SELECT name FROM sqlite_master WHERE type = 'table'");
        while ($table = $tableQuery->fetchArray(SQLITE3_NUM)) $tables[] = $table[0];
        if (!in_array('frc_user', $tables)) {
            $query = <<<'SQL'
                CREATE TABLE frc_user (
                    id       INTEGER NOT NULL PRIMARY KEY,
                    email    TEXT    NOT NULL,
                    password TEXT    NOT NULL)
                SQL;
            $db->exec($query);
            $db->exec('CREATE INDEX idx_user_email ON frc_user (email)');
        }
        if (!in_array('feed', $tables)) {
            $query = <<<'SQL'
                CREATE TABLE feed (
                    id         INTEGER NOT NULL PRIMARY KEY,
                    user_id    INTEGER NOT NULL,
                    url        TEXT    NOT NULL,
                    title      TEXT,
                    updated_on TEXT,
                    checked_on TEXT,
                    FOREIGN KEY (user_id) REFERENCES frc_user (id))
                SQL;
            $db->exec($query);
        }
        if (!in_array('item', $tables)) {
            $query = <<<'SQL'
                CREATE TABLE item (
                    id            INTEGER NOT NULL PRIMARY KEY,
                    feed_id       INTEGER NOT NULL,
                    title         TEXT    NOT NULL,
                    item_guid     TEXT    NOT NULL,
                    item_link     TEXT    NOT NULL,
                    published_on  TEXT    NOT NULL,
                    updated_on    TEXT,
                    content       TEXT    NOT NULL,
                    is_encoded    BOOLEAN NOT NULL,
                    is_read       BOOLEAN NOT NULL DEFAULT 0,
                    is_bookmarked BOOLEAN NOT NULL DEFAULT 0,
                    FOREIGN KEY (feed_id) REFERENCES feed (id))
                SQL;
            $db->exec($query);
        }
        $db->close();
    }

    /**
     * Find a user by their ID
     *
     * @param string $email The e-mail address of the user to retrieve
     * @return array|null The user information, or null if the user is not found
     */
    public static function findUserByEmail(string $email): ?array {
        $db = self::getConnection();
        $query = $db->prepare('SELECT * FROM frc_user WHERE email = :email');
        $query->bindValue(':email', $email);
        $result = $query->execute();
        if ($result) {
            $user = $result->fetchArray(SQLITE3_ASSOC);
            if ($user) return $user;
            return null;
        }
        return null;
    }

    /**
     * Add a user
     *
     * @param string $email The e-mail address for the user
     * @param string $password The user's password
     */
    public static function addUser(string $email, string $password): void {
        $db = self::getConnection();
        $query = $db->prepare('INSERT INTO frc_user (email, password) VALUES (:email, :password)');
        $query->bindValue(':email', $email);
        $query->bindValue(':password', password_hash($password, PASSWORD_DEFAULT));
        $query->execute();
    }

    /**
     * Parse/format a date/time from a string
     *
     * @param ?string $value The date/time to be parsed and formatted
     * @return string|null The date/time in `DateTimeInterface::ATOM` format, or `null` if the input cannot be parsed
     */
    private static function formatDate(?string $value): ?string {
        try {
            return $value ? (new DateTimeImmutable($value))->format(DateTimeInterface::ATOM) : null;
        } catch (Exception) {
            return null;
        }
    }

    /**
     * Add an RSS feed
     *
     * @param string $url The URL for the RSS feed
     * @param string $title The title of the RSS feed
     * @param ?string $updatedOn The date/time the RSS feed was last updated (from the XML, not when we checked)
     * @return int The ID of the added feed
     */
    public static function addFeed(string $url, string $title, ?string $updatedOn): int {
        $db = self::getConnection();
        $query = $db->prepare(<<<'SQL'
            INSERT INTO feed (
                user_id, url, title, updated_on, checked_on
            ) VALUES (
                :user, :url, :title, :updated, :checked
            )
            SQL);
        $query->bindValue(':user',    $_REQUEST['FRC_USER_ID']);
        $query->bindValue(':url',     $url);
        $query->bindValue(':title',   $title);
        $query->bindValue(':updated', self::formatDate($updatedOn));
        $query->bindValue(':checked', self::formatDate('now'));
        $result = $query->execute();
        return $result ? $db->lastInsertRowID() : -1;
    }

    /**
     * Does a feed item already exist?
     *
     * @param int $feedId The ID of the feed to which the item belongs
     * @param string $guid The GUID from the RSS feed, uniquely identifying the item
     * @return bool True if the item exists, false if not
     */
    public static function itemExists(int $feedId, string $guid): bool {
        $db = self::getConnection();
        $query = $db->prepare('SELECT COUNT(*) FROM item WHERE feed_id = :feed AND item_guid = :guid');
        $query->bindValue(':feed', $feedId);
        $query->bindValue(':guid', $guid);
        $result = $query->execute();
        return $result && $result->fetchArray(SQLITE3_NUM)[0] == 1;
    }

    /**
     * Add a feed item
     *
     * @param int $feedId The ID of the feed to which the item should be added
     * @param string $guid The GUID from the RSS feed (uses link if `<guid>` not specified)
     * @param string $link The link to this item
     * @param string $title The title of the item
     * @param string $publishedOn The date/time the item was published
     * @param ?string $updatedOn The date/time the item was last updated
     * @param string $content The content of the item
     * @param bool $isEncoded Whether the content has HTML (true) or is plaintext (false)
     */
    public static function addItem(int $feedId, string $guid, string $link, string $title, string $publishedOn,
                                   ?string $updatedOn, string $content, bool $isEncoded): void {
        $db = self::getConnection();
        $query = $db->prepare(<<<'SQL'
            INSERT INTO item (
                feed_id, item_guid, item_link, title, published_on, updated_on, content, is_encoded
            ) VALUES (
                :feed, :guid, :link, :title, :published, :updated, :content, :encoded
            )
            SQL);
        $query->bindValue(':feed',      $feedId);
        $query->bindValue(':guid',      $guid);
        $query->bindValue(':link',      $link);
        $query->bindValue(':title',     $title);
        $query->bindValue(':published', self::formatDate($publishedOn));
        $query->bindValue(':updated',   self::formatDate($updatedOn));
        $query->bindValue(':content',   $content);
        $query->bindValue(':encoded',   $isEncoded);
        $query->execute();
    }
}
