<?php
/**
 * Feed retrieval, parsing, and manipulation
 */
class Feed {

    /** @var string The XML namespace for Atom feeds */
    public const string ATOM_NS = 'http://www.w3.org/2005/Atom';

    /** @var string The XML namespace for the `<content:encoded>` tag that allows HTML content in a feed */
    public const string CONTENT_NS = 'http://purl.org/rss/1.0/modules/content/';

    /**
     * When parsing XML into a DOMDocument, errors are presented as warnings; this creates an exception for them
     *
     * @param int $errno The error level encountered
     * @param string $errstr The text of the error encountered
     * @return bool False, to delegate to the next error handler in the chain
     * @throws DOMException If the error is a warning
     */
    private static function xmlParseError(int $errno, string $errstr): bool {
        if ($errno == E_WARNING && substr_count($errstr, 'DOMDocument::loadXml()') > 0) {
            throw new DOMException($errstr, $errno);
        }
        return false;
    }

    /**
     * Parse a feed into an XML tree
     *
     * @param string $content The feed's RSS content
     * @return array|DOMDocument[]|string[] [ 'ok' => feed ] if successful, [ 'error' => message] if not
     */
    public static function parseFeed(string $content): array {
        set_error_handler(self::xmlParseError(...));
        try {
            $feed = new DOMDocument();
            $feed->loadXML($content);
            return [ 'ok' => $feed ];
        } catch (DOMException $ex) {
            return [ 'error' => $ex->getMessage() ];
        } finally {
            restore_error_handler();
        }
    }

    /**
     * Retrieve the feed
     *
     * @param string $url
     * @return array|DOMDocument[]|string[] [ 'ok' => feedXml, 'url' => actualUrl ] if successful,
     *      [ 'error' => message ] if not
     */
    public static function retrieveFeed(string $url): array {
        $feedReq = curl_init($url);
        curl_setopt($feedReq, CURLOPT_FOLLOWLOCATION, true);
        curl_setopt($feedReq, CURLOPT_RETURNTRANSFER, true);
        curl_setopt($feedReq, CURLOPT_CONNECTTIMEOUT, 5);
        curl_setopt($feedReq, CURLOPT_TIMEOUT,        15);

        $feedContent = curl_exec($feedReq);

        $result = array();
        $error  = curl_error($feedReq);
        $code   = curl_getinfo($feedReq, CURLINFO_RESPONSE_CODE);
        if ($error) {
            $result['error'] = $error;
        } else if ($code == 200) {
            $parsed = self::parseFeed($feedContent);
            if (array_key_exists('error', $parsed)) {
                $result['error'] = $parsed['error'];
            } else {
                $result['ok']  = $parsed['ok'];
                $result['url'] = curl_getinfo($feedReq, CURLINFO_EFFECTIVE_URL);
            }
        } else {
            $result['error'] = "HTTP Code $code: $feedContent";
        }

        curl_close($feedReq);
        return $result;
    }

    /**
     * Get the value of a child element by its tag name
     *
     * @param DOMElement $element The parent element
     * @param string $tagName The name of the tag whose value should be obtained
     * @return string The value of the element (or "[element] not found" if that element does not exist)
     */
    private static function eltValue(DOMElement $element, string $tagName): string {
        $tags = $element->getElementsByTagName($tagName);
        return $tags->length == 0 ? "$tagName not found" : $tags->item(0)->textContent;
    }

    /**
     * Update a feed's items
     *
     * @param int $feedId The ID of the feed to which these items belong
     * @param DOMElement $channel The RSS feed items
     * @return array [ 'ok' => true ] if successful, [ 'error' => message ] if not
     */
    public static function updateItems(int $feedId, DOMElement $channel): array {
        try {
            foreach ($channel->getElementsByTagName('item') as $item) {
                $itemGuid = self::eltValue($item, 'guid');
                if ($itemGuid == 'guid not found') $itemGuid = self::eltValue($item, 'link');
                $isNew = !Data::itemExists($feedId, $itemGuid);
                if ($isNew) {
                    $title     = self::eltValue($item, 'title');
                    $link      = self::eltValue($item, 'link');
                    $published = self::eltValue($item, 'pubDate');
                    $updNodes  = $item->getElementsByTagNameNS(self::ATOM_NS, 'updated');
                    $updated   = $updNodes->length > 0 ? $updNodes->item(0)->textContent : null;
                    $encNodes  = $item->getElementsByTagNameNS(self::CONTENT_NS, 'encoded');
                    if ($encNodes->length > 0) {
                        $content = $encNodes->item(0)->textContent;
                        $isEncoded = true;
                    } else {
                        $content = self::eltValue($item, 'description');
                        $isEncoded = false;
                    }
                    Data::addItem($feedId, $itemGuid, $link, $title, $published, $updated, $content, $isEncoded);
                } // TODO: else check updated date; may want to return that from the isNew check instead
            }
        } catch (Exception $ex) {
            return [ 'error' => $ex->getMessage() ];
        }
        return [ 'ok', true ];
    }

    /**
     * Add an RSS feed
     *
     * @param string $url The URL of the RSS feed to add
     * @return array [ 'ok' => true ] if successful, [ 'error' => message ] if not
     */
    public static function add(string $url): array {
        $feed = self::retrieveFeed($url);
        if (array_key_exists('error', $feed)) return $feed;

        $channel = $feed['ok']->getElementsByTagName('channel')->item(0);
        if (!$channel instanceof DOMElement) return [ 'error' => "Channel element not found ($channel->nodeType)" ];

        // In Atom feeds, lastBuildDate contains the last time an item in the feed was updated; if that is not present,
        // use the pubDate element instead
        $updated = self::eltValue($channel, 'lastBuildDate');
        if ($updated == 'lastBuildDate not found') {
            $updated = self::eltValue($channel, 'pubDate');
            if ($updated == 'pubDate not found') $updated = null;
        }
        $feedId = Data::addFeed($feed['url'], self::eltValue($channel, 'title'), $updated);

        $result = self::updateItems($feedId, $channel);
        if (array_key_exists('error', $result)) return $result;

        return [ 'ok' => true ];
    }
}
